// This code is lifted pretty much wholesale from Matt Reiferson's DJGPP
// Game Programming
// Tutorial.

// I removed the main, and declared the globals in INPUT.H as
// extern [whatever type they are] made function prototypes and so on
// - so that functions/variables could be manipulated and used from the
// outside...

// Also, I've added something new...

// my buddy Dhonn Lushine told me I should be locking my ISR code
// & data... and so for the key_handler - I've added a function
// key_handler_end... here's how it's used...

// _go32_dpmi_lock_data(&keys_active,sizeof(keys_active));
// _go32_dpmi_lock_code(key_handler,(long)key_handler_end-(long)key_handler);

// those two lines LOCK the code & data for the keyboard ISR...

// the same should be done with the mouse, joystick, & timer (if you
// make a new timer interrupt handler)...

// you need to lock code & data because hardware interrupts do allow paging,
// you don't want any of these routines to have to be "swapped" from the
// hard disk when you need them - lock 'em up

#include <dpmi.h>
#include <dos.h>
#include <go32.h>
#include <pc.h>
#include <stdio.h>
#include "input.h"

unsigned char volatile raw_key, key_table[128];

int keys_active=0;
	
_go32_dpmi_seginfo old_key_handler,new_key_handler;

/* joystick */
unsigned int joy_1_max_x, joy_1_max_y,
	joy_1_min_x, joy_1_min_y,
	joy_1_cx, joy_1_cy,
	joy_2_max_x, joy_2_max_y,
	joy_2_min_x, joy_2_min_y,
	joy_2_cx, joy_2_cy;


unsigned char joystick_button(unsigned char button)
{
	outportb(JOYPORT,0); /* clear the latch and request a sample */

	/* invert button then and with request */
	return( ~inportb(JOYPORT) & button);
}

int joystick_pos(int *x, int *y)
{
        unsigned char portvalue;

	(*x) = (*y) = 0;
	asm("cli");
	outportb(JOYPORT,0); /* clear the latch and request a sample */
	do {
		portvalue = inportb(JOYPORT);
		if (portvalue & J1_X) {
			(*x)++;
		}
		if (portvalue & J1_Y) {
		       (*y)++;
		}
	} while ((portvalue & 0x03) && ((*y) < 0xffff) && ((*x) < 0xffff));
	asm("sti");

	if ((*x >= 0xffff) || (*y >= 0xffff)) return 0; /* timout */

	return 1;
}

unsigned int detect_joystick(void)
{
	int x, y;

	return joystick_pos(&x,&y);
}

void joystick_calibrate(void)
{
	unsigned int x_new,y_new; // temp joystick positions

	joy_1_min_x = joy_1_min_y = 0xffff;
	joy_1_max_x = joy_1_max_y = 0;

	printf("Calibrating Joystick\n");

	printf("Swirl the joystick and press a button\n");
	do {
		joystick_pos(&x_new,&y_new);
		if (x_new < joy_1_min_x) joy_1_min_x = x_new;
		if (x_new > joy_1_max_x) joy_1_max_x = x_new;
		if (y_new < joy_1_min_y) joy_1_min_y = y_new;
		if (y_new > joy_1_max_y) joy_1_max_y = y_new;
	} while(!joystick_button(J1_A | J1_B));

	while(joystick_button(J1_A | J1_B)) {
	}

	printf("Center the joystick and press a button\n");
	while(!joystick_button(J1_A | J1_B)) {
		joystick_pos(&joy_1_cx,&joy_2_cx);
	}

	while(joystick_button(J1_A | J1_B)) {
	}
}

void key_handler(void)
{
        unsigned char al;

	asm("cli; pusha");
	
	raw_key = inportb(0x60);
		
	al = inportb(0x61); 
	al |= 0x82;
	outportb(0x61, al);       
	al &= 0x7f;
	outportb(0x61, al);

	/* you have the option of putting this outside */
	if(raw_key < 128) {
         if(!key_table[raw_key])
            {
     	      key_table[raw_key] = 1;
            keys_active++;
            }
	}
      else {
         if(key_table[raw_key-128])
            {
            key_table[raw_key-128] = 0;
            keys_active--;
            }
      }

	outportb(0x20, 0x20);
	asm("popa; sti");
}

void key_handler_end(void){}

void key_init(void) /* function to swap state */
{
      int index;
	new_key_handler.pm_offset   = (int)key_handler;
	new_key_handler.pm_selector = _go32_my_cs();
	_go32_dpmi_get_protected_mode_interrupt_vector(0x9, &old_key_handler);
	_go32_dpmi_allocate_iret_wrapper(&new_key_handler);
	_go32_dpmi_set_protected_mode_interrupt_vector(0x9,&new_key_handler);
      for(index=0;index<128;index++)
         key_table[index] = 0;
}

void key_delete(void)
{
	_go32_dpmi_set_protected_mode_interrupt_vector(0x9,&old_key_handler);
}

int mouse_detect(void)
{
	union REGS regs;

	regs.x.ax = 0x0;
	int86(0x33, &regs, &regs);

	return regs.x.ax;
}

void mouse_show(void)
{
	union REGS regs;

	regs.x.ax = 0x01;
	int86(0x33, &regs, &regs);
}

void mouse_hide(void)
{
	union REGS regs;

	regs.x.ax = 0x02;
	int86(0x33, &regs, &regs);
}

unsigned char mouse_button(unsigned char button)
{
	union REGS regs;

	regs.x.ax = 0x3;
	int86(0x33, &regs, &regs);

	return (regs.x.bx & button);
}

void mouse_abs_pos(int *x, int *y)
{
	union REGS regs;

	regs.x.ax = 0x03;
	int86(0x33, &regs, &regs);
	*x = regs.x.cx;
	*y = regs.x.dx;

	return;
}

int mouse_rel_pos(int *x, int *y)
{
	union REGS regs;

	regs.x.ax = 0xb;
	int86(0x33, &regs, &regs);
	*x = regs.x.cx;
	*y = regs.x.dx;

	return (*x|*y);
}


