// This file Modey.cc has many functions that were made by me: James McCue
// NOTE: All Sprite routines assume a width evenly divisible by 4 - remember
//  that!
// Author's homepage:
//       http://www.fortunecity.com/skyscraper/gigahertz/179/index.htm
//
// Some code is from (or modified from):
//    Andre' Lamothe's Black Art of 3D Game Programming (waite group press)
//    Dave Roberts' PC Game Programming Explorer (coriolis group)
//    Michael Abrash's Zen of Graphics Programming (coriolis group)
//
//  - and I'd like to take the opportunity to say that this about some of
// my homebrew routines: THEY STINK!
//
// I mean... analysis of many of the functions will show you that there
// are many instances where work is being done in the inner loops that could
// be - in part - taken out of the inner loops...
// As my work with mode Y involved the barest minimum of functions:
//    Setting the mode
//    Page Flipping
//    Filling the screen (or an area of it) with a color
//    Drawing vertical textured strips
//
//  ...and the fact I was using "linear sprites" for said textured strips
// - I didn't have much need to delve into the benefits of planar bitmaps,
// and READING the video memory...
//
// This module though - does those things - and I've tried to make clear
// what is going on in each function... The Sprite manipulation functions
// assume you're using page 0 and page 1 as a pageflipping tandem...
//
// It's been a learning experience for me... writing these "tuts".
//
// I hope you can benefit by them!
//
// This source is without Warrantee of any kind. James McCue disclaims all
// warrantees as far as the functionality of this source code.
//
// James McCue is not responsible for any damages that may arise from the
// use or abuse of this code.

#include <sys/nearptr.h>
#include <go32.h>
#include <dos.h>         
#include <stdlib.h>
#include <stdio.h>
#include <pc.h>
#include <dpmi.h>
#include <time.h>
#include "modez.h"

char *video_buffer = (char *)0xa0000;
char *double_buffer;
char *rom_char_set = (char *)0xffa6e;


char *pagez_0_buffer  = (char *)0xa0000;
char *pagez_1_buffer  = (char *)0xa8000;

int mode_z_page                   = PAGE_0;

int screen_width_mode;
int screen_height_mode;

void Fill_Screen_Z(int color)
{
// this function will fill the mode Z video buffer with the sent color
   long new_color;

   new_color = color | (color<<8);
   new_color = new_color | (new_color<<16);

   outportb(SEQUENCER,SEQ_PLANE_ENABLE);
   outportb(SEQUENCER+1,0x0f);

// use the inline assembler macro for speed

   rep_stosl(new_color, video_buffer, 8000);
} // end Fill_Screen_Z

///////////////////////////////////////////////////////////////////////////////

void Write_Pixel_Z(int x,int y,int color)
{

// this function will write a pixel to screen in mode Y

// first select the proper color plane ( actually - in Mode Y - we're talking
// pixel planes... hehe... but remember:
//   We're using hardware of the VGA that was initially created for use
// with the 16 color modes - where planes overlap.

   int what_plane;
   int for_shifting;

   what_plane = x & 0x03;
   for_shifting = 0x01;
   for_shifting = for_shifting<<what_plane;

   outportb(SEQUENCER,SEQ_PLANE_ENABLE);
   outportb(SEQUENCER+1,for_shifting);

// write the pixel, offset = (y*320+x)/4

   video_buffer[(y<<6)+(y<<4)+(x>>2)] = (unsigned char )color;

} // end Write_Pixel_Z

void Set_Mode_Z(void)
{
// this function will set the video mode to 320x200x256

int data;  // used to store data
__dpmi_regs regs;

// set system to mode 13h and use it as a foundation to base 320x200 mode on

memset(&regs, 0, sizeof regs);    // is this memset really necessary???
regs.x.ax = 0x13; /* 0x13 is the mode number */
__dpmi_int(0x10, &regs);


// make changes to the crt controller first

outportb(CRT_CONTROLLER, CRT_MAX_SCANLINE);
data=inportb(CRT_CONTROLLER+1);
outportb(CRT_CONTROLLER+1, RESET_BITS(data,0x0f));

// use byte addressing instead of word

outportb(CRT_CONTROLLER,CRT_ADDR_MODE);
data=inportb(CRT_CONTROLLER+1);
outportb(CRT_CONTROLLER+1,RESET_BITS(data,0x40));

// second register that needs to reflect byte addressing

outportb(CRT_CONTROLLER,CRT_MODE_CONTROL);
data=inportb(CRT_CONTROLLER+1);
outportb(CRT_CONTROLLER+1,SET_BITS(data,0x40));

// make changes to graphics controller

// set addressing to not use odd/even memory writes

outportb(GFX_CONTROLLER,GFX_WRITE_MODE);
data=inportb(GFX_CONTROLLER+1);
outportb(GFX_CONTROLLER+1,RESET_BITS(data,0x10));

// don't chain the memory maps together

outportb(GFX_CONTROLLER,GFX_MISC);
data=inportb(GFX_CONTROLLER+1);
outportb(GFX_CONTROLLER+1,RESET_BITS(data,0x02));

// make changes to sequencer

// again we must select no chaining and no odd/even memory addressing

outportb(SEQUENCER,SEQ_MEMORY_MODE);
data =inportb(SEQUENCER+1);
data = RESET_BITS(data,0x08);
data = SET_BITS(data,0x04);
outportb(SEQUENCER+1,data);

// now clear the screen

outportb(SEQUENCER,SEQ_PLANE_ENABLE);
outportb(SEQUENCER+1,0x0f);

// clear the screen, remember it is 320x200, but that is divided into four
// planes, hence we need only to clear 16k out since there will be four planes
// each being cleared in parallel for a total of 4*16k or 64 = 320x200
// note: "k" in this example means 1000 not 1024

// clear BOTH pages

rep_stosl(0x00,video_buffer,16000);

// set screen width/height variables

screen_width_mode = 80;
screen_height_mode = 400;

} // end Set_Mode_Z

// The following functions are purely HIGH-JACKED CODE.

void Wait_Vertical_Retrace_End(void)
{
    unsigned char InputStatus1;

    /* make sure we aren't in display state (= not retrace) */
    do {
    	InputStatus1 = inportb(VGA_INPUT_STATUS_1);
    } while (!(InputStatus1 & VR_BIT));

    /* wait for retrace to end */
    do {
    	InputStatus1 = inportb(VGA_INPUT_STATUS_1);
    } while (InputStatus1 & VR_BIT);
}

/*
	Function: WaitDisplayMode
    Description:
    	If the VGA is currently in display mode, the function returns
        immediately, else it waits until diplay mode is entered.
*/
void Wait_Display_Mode(void)
{
	unsigned char InputStatus1;

    /* wait for retrace to end and display mode to start */
    do
    {
    	InputStatus1 = inportb(VGA_INPUT_STATUS_1);
    } while (InputStatus1 & DISPLAY_ENABLE);
}

/*
	Function: WaitRetraceMode
    Description:
    	If the VGA is currently in retrace mode, the function returns
        immediately, else it waits until retrace mode is entered.
*/
void Wait_Retrace_Mode(void)
{
	unsigned char InputStatus1;

    /* wait for display mode to end and retrace to start */
    do
    {
    	InputStatus1 = inportb(VGA_INPUT_STATUS_1);
    } while (!(InputStatus1 & VR_BIT));
}

/*
	Function: VerticalRetraceOccuring
    Description:
    	Returns the current state of the VGA vertical retrace.  If
        retrace is occuring, the function returns TRUE.
*/
int Vertical_Retrace_Occuring(void)
{
	return (inportb(VGA_INPUT_STATUS_1) & VR_BIT) >> 3;
}


void Set_Visual_Page_Mode_Z(int page)
{
// this function sets the visual page that will be displayed by the VGA

// disable interrupts...

__asm__("cli\n\t");

if (page==PAGE_0)
   {
   // re-program the start address registers in the CRT controller
   // to point at page 0 @ 0xA000:0000

   // first low byte of address

   outportb(CRT_CONTROLLER,CRT_ADDR_LOW);
   outportb(CRT_CONTROLLER+1,0x00);

   // now high byte

   outportb(CRT_CONTROLLER,CRT_ADDR_HI);
   outportb(CRT_CONTROLLER+1,0x00);

   } // end if page 0
else
if (page==PAGE_1)
   {
   // re-program the start address registers in the CRT controller
   // to point at page 1 @ 0xA000:8000

   // first low byte of address

   outportb(CRT_CONTROLLER,CRT_ADDR_LOW);
   outportb(CRT_CONTROLLER+1,0x00);

   // now high byte

   outportb(CRT_CONTROLLER,CRT_ADDR_HI);
   outportb(CRT_CONTROLLER+1,0x80);

   } // end else page 1

// note: we could use WORD out's, but this is clearer, feel free to change them

   // re-enable interrupts
__asm__("sti\n\t");


} // end Set_Visual_Page_Mode_Z

///////////////////////////////////////////////////////////////////////////////

void Set_Visual_Page_Mode_Z_Half(int page)
{
// THIS VERSION OF THE FUNCTION SAVES AN OUTS WORTH OF COMPUTER CYCLES!!!

// this function sets the visual page that will be displayed by the VGA

// disable interrupts

__asm__("cli\n\t");


if (page==PAGE_0)
   {
   // re-program the start address registers in the CRT controller
   // to point at page 0 @ 0xA000:0000

   // now high byte

   outportb(CRT_CONTROLLER,CRT_ADDR_HI);
   outportb(CRT_CONTROLLER+1,0x00);

   } // end if page 0
else
if (page==PAGE_1)
   {
   // re-program the start address registers in the CRT controller
   // to point at page 1 @ 0xA000:8000

   // now high byte

   outportb(CRT_CONTROLLER,CRT_ADDR_HI);
   outportb(CRT_CONTROLLER+1,0x80);

   } // end else page 1

// re-enable interrupts

__asm__("cli\n\t");


} // end Set_Visual_Page_Mode_Z_Half

///////////////////////////////////////////////////////////////////////////////

void Set_Working_Page_Mode_Z(int page)
{
// this function sets the page that all MODE Y functions will update when
// called

if (page==PAGE_0)
   video_buffer = pagez_0_buffer;
else
if (page==PAGE_1)
   video_buffer = pagez_1_buffer;

} // end Set_Working_Page_Mode_Z

void PCX_Copy_To_Video_Mem(pcx_picture_ptr image, unsigned char *dest)
{
// This function copies a PCX PICTURE into a region of VGA memory

    int x_index;            // looping variable
    int y_index;            // looping variable

    int what_plane;
    int for_shifting;

    int next_image_row_offset = 0;
    int next_video_line_offset = 0;
    int bottom = 0;

    unsigned int image_buffer_offset; // offset into image to be copied
    unsigned int vid_offset;          // offset into never displayed memory

    char *the_image;

    next_image_row_offset = image->header.width - 320;

    if (image->header.width > 400) bottom = 400;
    else bottom = 400;

    the_image = image->buffer;

    for (x_index = 0; x_index<320; x_index++)
    {
    // first select the proper color plane to write to...
    // I could have written to one plane at a time but - Oh Hell!...

        what_plane = x_index & 0x03;
        for_shifting = 0x01;
        for_shifting = for_shifting<<what_plane;

	outportb(SEQUENCER,SEQ_PLANE_ENABLE);
	outportb(SEQUENCER+1,for_shifting);

    	for (y_index = 0; y_index<bottom; y_index++)
    	{
            image_buffer_offset = (y_index<<8) + (y_index<<6) + x_index+(y_index*next_image_row_offset);

            vid_offset = (y_index<<6) + (y_index<<4) + (x_index>>2);

            dest[vid_offset] =the_image[image_buffer_offset];
    	}
    }
} // end PCX_Copy_To_Video_Mem

void Copy_From_Unseen_Vram_to_Page(unsigned char *buffer1, unsigned char *buffer2,int to_row)
{
// this function copies a [mode Y] screen full 'o data from the area of
// memory pointed to by 'buffer1', and copies it to the page pointed to by
// 'buffer2' (usually PAGE 0 or PAGE 1)...

// THIS IS MY FIRST REAL   W R I T E   M O D E   O N E 	experiment!

// The function uses WRITE MODE ONE to copy all four planes at ONCE!

    unsigned int ModeRegTemp;

    unsigned char pixel_color;

    unsigned char *buff2;
    unsigned char *buff1;
    int xw,yw;

    // Write Mode ONE...

    outportb(GFX_CONTROLLER, GFX_WRITE_MODE);

    ModeRegTemp = inportb(GC_DATA_REG);
    ModeRegTemp = (ModeRegTemp & 0xFC) | 0x01;
    outportb(GC_DATA_REG, ModeRegTemp);

	// ... and the stuff

    buff1 = buffer1;
    buff2 = buffer2;

    // enable all planes

    outportb(SEQUENCER,SEQ_PLANE_ENABLE);
    outportb(SEQUENCER+1,0x0f);

    for (yw = 0; yw<to_row; yw++)
    {
        for (xw = 0; xw < screen_width_mode; xw++)
        {
            *buff2=*buff1;
            buff2++;
            buff1++;
        }
    }

    // Now... leave WRITE MODE ONE!

    ModeRegTemp &= 0xFC;

    outportb(GFX_CONTROLLER, GFX_WRITE_MODE);
    outportb(GC_DATA_REG, ModeRegTemp);

} // end Copy_From_Unseen_Vram_to_Page

void Planar_Special_Fade_to(unsigned char *buffer1, unsigned char *buffer2, int to_row)
{
// this function copies a [mode Y] screen full 'o data from the area of
// memory pointed to by 'buffer1', and copies it to the page point to by
// 'buffer2' (usually PAGE 0 or PAGE 1)...

// This routine will do it one plane at a time while you watch... if you
// want it to...

// THIS IS MY SECOND REAL   W R I T E    M O D E   O N E experiment!

    unsigned int ModeRegTemp;

    unsigned char pixel_color;

    unsigned char *buff2;
    unsigned char *buff1;
    int xw,yw;

    // Write Mode ONE...

    outportb(GFX_CONTROLLER, GFX_WRITE_MODE);

    ModeRegTemp = inportb(GC_DATA_REG);
    ModeRegTemp = (ModeRegTemp & 0xFC) | 0x01;
    outportb(GC_DATA_REG, ModeRegTemp);

    // ... and the stuff

    buff1 = buffer1;
    buff2 = buffer2;

    outportb(SEQUENCER,SEQ_PLANE_ENABLE);
    outportb(SEQUENCER+1,0x08);

    for (yw = 0; yw<to_row; yw++)
    {
        for (xw = 0; xw < screen_width_mode; xw++)
        {
            *buff2=*buff1;
            buff2++;
            buff1++;
        }
    }

    buff1 = buffer1;
    buff2 = buffer2;

    outportb(SEQUENCER,SEQ_PLANE_ENABLE);
    outportb(SEQUENCER+1,0x04);

    for (yw = 0; yw<to_row; yw++)
    {
        for (xw = 0; xw < screen_width_mode; xw++)
        {
            *buff2=*buff1;
            buff2++;
            buff1++;

        }
    }

    buff1 = buffer1;
    buff2 = buffer2;

    outportb(SEQUENCER,SEQ_PLANE_ENABLE);
    outportb(SEQUENCER+1,0x02);

    for (yw = 0; yw<to_row; yw++)
    {
        for (xw = 0; xw < screen_width_mode; xw++)
        {
            *buff2=*buff1;
            buff2++;
            buff1++;

        }
    }

    buff1 = buffer1;
    buff2 = buffer2;

    outportb(SEQUENCER,SEQ_PLANE_ENABLE);
    outportb(SEQUENCER+1,0x01);

    for (yw = 0; yw<to_row; yw++)
    {
        for (xw = 0; xw < screen_width_mode; xw++)
        {
            *buff2=*buff1;
            buff2++;
            buff1++;

        }
    }

    // Now... leave WRITE MODE ONE!

    ModeRegTemp &= 0xFC;

    outportb(GFX_CONTROLLER, GFX_WRITE_MODE);
    outportb(GC_DATA_REG, ModeRegTemp);

} // end Planar_Special_Fade_to

//***********************************************************************
// WARNING: The Function Print_Char (although it works) is GLARINGLY
// inefficent and confusing... check back in at:
//  http://www.perlexed.com/GPMega
//  -because in future versions of the tutorial... we'll work on printing
// strings in Mode Y, explain it - and give you MUCH better code to play
// with!
//***********************************************************************

void Print_Char_Mode_Z(int xc,int yc,char c,int color,int transparent)
{
// this function is used to print a character on the screen. It uses the
// internal 8x8 character set to do this. Note that each character is
// 8 bytes where each byte represents the 8 pixels that make up the row
// of pixels

// This function cuts down on OUTS by 'unrolling the loop' and printing
// a character every FOURTH row at a time.

// Many times... if you compare mode x functions to mode13h functions
// you'll notice the loops are reversed...

// in mode13h... you'll find many things are set up so that you can do
// memcopies or memsets, along horizontal lines..

// since Mode Y requires OUTS whenever you step over to the next pixel,
// every effort is made to use VERTICAL scan conversion.

// It's like that 90 degree turn we all keep hearing about!

int offset,             // offset into video memory
    x,               	// loop variable
    y;               	// loop variable
    int new_x,new_y;
    unsigned char *work_char; 	// pointer to character being printed

    unsigned char bit_mask;       	// bitmask used to extract proper bit

    int what_plane;
    int for_shifting;

    // compute starting offset in rom character lookup table
    // multiple the character by 8 and add the result to the starting address
    // of the ROM character set

    // compute offset of character in video buffer, use shifting to multiply
    // draw the character row by row

    // set bit mask

    bit_mask = 0x80;

    // draw each character by columns... this saves OUTS

    for (x=0; x<ROM_CHAR_WIDTH; x++)
    {
        new_x = xc+x;
        what_plane = (new_x) & 0x03;
        for_shifting = 0x01;
        for_shifting = for_shifting<<what_plane;

        outportb(SEQUENCER,SEQ_PLANE_ENABLE);
        outportb(SEQUENCER+1,for_shifting);

        work_char = rom_char_set + c * ROM_CHAR_HEIGHT;


        for (y=0; y<ROM_CHAR_HEIGHT; y++)
        {
            new_y = yc+y;
            offset = (new_y << 6) + (new_y << 4) + (new_x>>2);

            // test for transparent pixel i.e. 0, if not transparent then draw


            if ((*work_char & bit_mask))
                video_buffer[offset] = (unsigned char)color;
            else
            if (!transparent)               // takes care of transparency
                video_buffer[offset] = 0; // make black part opaque

    	    // move to next line in rom character data area
            work_char++;

        } // end for y

        // shift the bit mask
        bit_mask = (bit_mask>>1);

    } // end for x


} // end Print_Char_Mode_Z

//////////////////////////////////////////////////////////////////////////////

void Print_String_Mode_Z(int x,int y,int color, char *string,int transparent)
{
// this MODE Y function prints an entire string on the screen with fixed
// spacing between each character by calling the Print_Char_Mode_Z() function

 int index,   // loop index
     length;  // length of string

// compute length of string

length = strlen(string);

// print the string a character at a time

for (index=0; index<length; index++)
	  Print_Char_Mode_Z(x+(index<<3),y,string[index],color,transparent);

} // end Print_String_Mode_Z


void Line_H_Mode_Z(int x1,int x2,int y,int color)
{
// draw a horizontal line by "unrolling the loop"... drawing each plane of
// the horizontal line... one at a time.
// this function does not do clipping hence x1,x2 and y must all be within
// the bounds of the screen

// This function is a perfect example of the horrors of being away from
// directed study for to long... so much extraneous work can be ripped out
// here.

int index;
int temp; // used for temporary storage during endpoint swap

int what_plane;
int for_shifting;


// draw the row of pixels

   what_plane = x1 & 0x03;
   for_shifting = 0x01;
   for_shifting = for_shifting<<what_plane;

   outportb(SEQUENCER,SEQ_PLANE_ENABLE);
   outportb(SEQUENCER+1,for_shifting);

   for (index = x1; index <= x2; index+=4)
   {
        memset((char *)(video_buffer + ((y<<6) + (y<<4)) + (index>>2)),
	     		(unsigned char)color,1);

   }

   x1++;

   what_plane = x1 & 0x03;
   for_shifting = 0x01;
   for_shifting = for_shifting<<what_plane;

   outportb(SEQUENCER,SEQ_PLANE_ENABLE);
   outportb(SEQUENCER+1,for_shifting);

   for (index = x1; index <= x2; index+=4)
   {
        memset((char *)(video_buffer + ((y<<6) + (y<<4)) + (index>>2)),
	     		(unsigned char)color,1);

   }

   x1++;

   what_plane = x1 & 0x03;
   for_shifting = 0x01;
   for_shifting = for_shifting<<what_plane;

   outportb(SEQUENCER,SEQ_PLANE_ENABLE);
   outportb(SEQUENCER+1,for_shifting);


   for (index = x1; index <= x2; index+=4)
   {
       memset((char *)(video_buffer + ((y<<6) + (y<<4)) + (index>>2)),
	     		(unsigned char)color,1);

   }

   x1++;

   what_plane = x1 & 0x03;
   for_shifting = 0x01;
   for_shifting = for_shifting<<what_plane;

   outportb(SEQUENCER,SEQ_PLANE_ENABLE);
   outportb(SEQUENCER+1,for_shifting);

   for (index = x1; index <= x2; index+=4)
   {
       memset((char *)(video_buffer + ((y<<6) + (y<<4)) + (index>>2)),
	     		(unsigned char)color,1);

   }

} // end Line_H_Mode_Z


void Line_V_Mode_Z(int y1,int y2,int x,int color)
{
// draw a vertical line, note that a memset function can no longer be
// used since the pixel addresses are no longer contiguous in memory
// note that the end points of the line must be on the screen

unsigned char *start_offset; // starting memory offset of line

int index, // loop index
    temp;  // used for temporary storage during swap

int what_plane;
int for_shifting;


// make sure y2 > y1

if (y1>y2)
   {
   temp = y1;
   y1   = y2;
   y2   = temp;
   } // end swap

// compute starting position

   what_plane = x & 0x03;
   for_shifting = 0x01;
   for_shifting = for_shifting<<what_plane;

   outportb(SEQUENCER,SEQ_PLANE_ENABLE);
   outportb(SEQUENCER+1,for_shifting);

start_offset = video_buffer + ((y1<<6) + (y1<<4)) + (x>>2);

for (index=0; index<=y2-y1; index++)
    {
    // set the pixel

    *start_offset = (unsigned char)color;

    // move downward to next line

    start_offset+=screen_width_mode;

    } // end for index

} // end Line_V_Mode_Z

///////////////////////////////////////////

/* The following routines are the same general routines I used in
  mode13h as well...

   There isn't any "planar" sprite stuff in here... just the basic pcx
  loading and palette manipulation stuff... */

///////////////////////////////////////////

void Set_Palette_Register(int index, rgb_color_ptr color)
{
        outportb(PALETTE_MASK,0xff);
        outportb(PALETTE_REGISTER_WR,index);
        outportb(PALETTE_DATA,color->red);
        outportb(PALETTE_DATA,color->green);
        outportb(PALETTE_DATA,color->blue);
}

///////////////////////////////////////////////////////////////////////////////

void Get_Palette_Register(int index, rgb_color_ptr color)
{
        outportb(PALETTE_MASK,0xff);
        outportb(PALETTE_REGISTER_RD,index);
        color->red = inportb(PALETTE_DATA);
        color->green = inportb(PALETTE_DATA);
        color->blue = inportb(PALETTE_DATA);
}

/////////////////////////////////////////////////////////////////////////////

void PCX_Init(pcx_picture_ptr image, int width, int height)
{
// this function allocates the buffer region needed to load a pcx file
// NOTE: This code assumes 320x200 PCX images... it's a shortcoming...

if (!(image->buffer = (char *)malloc(width * height + 1)))

   printf("\ncouldn't allocate screen buffer");

image->header.width = width;
image->header.height= height;
} // end PCX_Init

//////////////////////////////////////////////////////////////////////////////

void PCX_Load(char *filename, pcx_picture_ptr image, int width, int height, int enable_palette)
{
// this function loads a pcx file into a picture structure, the actual image
// data for the pcx file is decompressed and expanded into a secondary buffer
// within the picture structure, the separate images can be grabbed from this
// buffer later.  also the header and palette are loaded

// NOTE: As with the procedure PCX_Init... the size of the PCX to be loaded
// is assumed to be 320x200... not very flexible...

FILE *fp;
int num_bytes,index;
long count;
unsigned char data;
char *temp_buffer;
int pic_width, pic_height;
// open the file

fp = fopen(filename,"rb");

// load the header

temp_buffer = (char *)image;

for (index=0; index<128; index++)
    {
    temp_buffer[index] = (char)getc(fp);
    } // end for index

// load the data and decompress into buffer

pic_width = width;
pic_height = height;

image->header.width = pic_width;
image->header.height = pic_height;

if (!(image->buffer = (char *)malloc(pic_width * pic_height + 1)))
{
   printf("\ncouldn't allocate screen buffer");
   return;
}

// load the data and decompress into buffer

count=0;

while(count<=pic_width * pic_height)
     {
     // get the first piece of data

     data = (unsigned char)getc(fp);

     // is this a rle?

     if (data>=192 && data<=255)
        {
        // how many bytes in run?

        num_bytes = data-192;

        // get the actual data for the run

        data  = (unsigned char)getc(fp);

        // replicate data in buffer num_bytes times

        while(num_bytes-->0)
             {
             image->buffer[count++] = data;

             } // end while

        } // end if rle
     else
        {
        // actual data, just copy it into buffer at next location

        image->buffer[count++] = data;

        } // end else not rle

     } // end while

// move to end of file then back up 768 bytes i.e. to begining of palette

fseek(fp,-768L,SEEK_END);

// load the pallete into the palette

for (index=0; index<256; index++)
    {
    // get the red component

    image->palette[index].red   = (unsigned char)(getc(fp) >> 2);

    // get the green component

    image->palette[index].green = (unsigned char)(getc(fp) >> 2);

    // get the blue component

    image->palette[index].blue  = (unsigned char)(getc(fp) >> 2);

    } // end for index

fclose(fp);

// change the palette to newly loaded palette if commanded to do so

if (enable_palette)
   {

   // for each palette register set to the new color values

   for (index=0; index<256; index++)
       {

       Set_Palette_Register(index,(rgb_color_ptr)&image->palette[index]);

       } // end for index

   } // end if change palette

} // end PCX_Load

void PCX_Delete(pcx_picture_ptr image)
{
// this function de-allocates the buffer region used for the pcx file load

free(image->buffer);

} // end PCX_Delete

void PCX_Get_Sprite(pcx_picture_ptr image,
                     sprite_ptr sprite,
                     int sprite_frame,
                     int cell_x, int cell_y)

{
// this function is used to load the images for a sprite into the sprite
// frames array. It functions by using the size of the sprite and the
// position of the requested cell to compute the proper location in the
// pcx image buffer to extract the data from.

int x_off,  // position of sprite cell in PCX image buffer
    y_off,
    y,      // looping variable
    width,  // size of sprite
    height;

unsigned char  *sprite_data;

// extract width and height of sprite

    width  = sprite->width;
    height = sprite->height;

    // first allocate the memory for the sprite in the sprite structure

    sprite->frames[sprite_frame] = (unsigned char  *)malloc(width * height + 1);

    // create an alias to the sprite frame for ease of access

    sprite_data = sprite->frames[sprite_frame];

    // now load the sprite data into the sprite frame array from the pcx picture

    x_off = (width+1)  * cell_x + 1;
    y_off = (height+1) * cell_y + 1;

    // compute starting y address

    y_off = y_off * 320; // 320 bytes per line

    // scan the data row by row

    for (y=0; y<height; y++,y_off+=320)
    {
        // copy the row of pixels

        memcpy((void  *)&sprite_data[y*width],

             (void  *)&(image->buffer[y_off + x_off]),
             width);

    } // end for y

    // increment number of frames

    sprite->num_frames++;

    // done!, let's bail!

} // end PCX_Get_Sprite

//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

void Sprite_Init(sprite_ptr sprite,int x,int y,int width,int height,
                                   int c1,int c2,int c3,
                                   int t1,int t2,int t3)
{
// this function initializes a sprite

int index;

sprite->x = x;
sprite->y = y;
sprite->width        = width;
sprite->height       = height;
sprite->visible      = 1;
sprite->counter_1    = c1;
sprite->counter_2    = c2;
sprite->counter_3    = c3;
sprite->threshold_1  = t1;
sprite->threshold_2  = t2;
sprite->threshold_3  = t3;
sprite->curr_frame   = 0;
sprite->state        = SPRITE_DEAD;
sprite->num_frames   = 0;
sprite->background   = (unsigned char  *)malloc(width * height+1);

// set all bitmap pointers to null

for (index=0; index<MAX_SPRITE_FRAMES; index++)
    sprite->frames[index] = NULL;

} // end Sprite_Init

//////////////////////////////////////////////////////////////////////////////

void Sprite_Delete(sprite_ptr sprite)
{
// this function deletes all the memory associated with a sprite

int index;

free(sprite->background);

// now de-allocate all the animation frames

for (index=0; index<MAX_SPRITE_FRAMES; index++)
    free(sprite->frames[index]);

} // end Sprite_Delete

// The following routine splits the screen...
// The lines FOLLOWING the split line will be scanned
// from the top of the "read" page 0... at
// a0000....
// This routine is another great find from Dave Roberts'
// PC Game Programming Explorer.

// We're altering the CRT CONTROLLERS line compare, overflow, and
// max scanline registers

void Set_Split_Screen(int line_number)
{
   // NOTE: Since modes X and Y are double scanned
   // (i.e.: truely possessing TWICE the video scan lines -
   //  mode X 320x240 is really 320x480, and mode Y 320x200 is
   // REALLY 320x400) the line number will be doubled here.

   // I have not tested this with mode z

   unsigned char Vga_temp;

   // sending 0xFFFF to this routine will effectively TURN OFF
   // our split screen

   if (line_number!=0xFFFF)
       line_number *= 2;

   // disable interrupts

   __asm__("cli\n\t");

   // address the CRT line compare register

   outportb(CRT_CONTROLLER, CRT_LINE_COMPARE);

   // write new value to the CRT Line Compare's DATA register

   outportb(CRT_CONTROLLER+1,line_number & 0xFF);

   // address the CRT overflow register

   outportb(CRT_CONTROLLER, CRT_OVERFLOW);

   // read the value from the CRT's overflow DATA register
   // alter it as needed...

   Vga_temp=inportb(CRT_CONTROLLER+1);
   Vga_temp=(Vga_temp & 0xEF) | ((line_number>>4) & 0x10);

   // write the new value to the CRT overflow's DATA register

   outportb(CRT_CONTROLLER+1, Vga_temp);

   // address the CRT max scanline's index register

   outportb(CRT_CONTROLLER, CRT_MAX_SCANLINE);

   // read the value from the CRT max scanline's DATA register
   // alter it as needed...

   Vga_temp=inportb(CRT_CONTROLLER+1);
   Vga_temp=(Vga_temp & 0xBF) | ((line_number>>3) & 0x40);

   // write the new value to the CRT max scanline's DATA register

   outportb(CRT_CONTROLLER+1, Vga_temp);

   // re-enable interrupts

   __asm__("sti\n\t");
} // end Set_Split_Screen
