// header file for ModeY.CC

// first... 2 memory moving/setting macros

#define rep_movsl(src, dest, numwords) \
__asm__ __volatile__ ( \
  "cld\n\t" \
  "rep\n\t" \
  "movsl" \
  : : "S" (src), "D" (dest), "c" (numwords) \
  : "%ecx", "%esi", "%edi" )

#define rep_stosl(value, dest, numwords) \
__asm__ __volatile__ ( \
  "cld\n\t" \
  "rep\n\t" \
  "stosl" \
  : : "a" (value), "D" (dest), "c" (numwords) \
  : "%ecx", "%edi" )

#define GRAPHICS        0x013
#define TEXT            0x03

#define PALETTE_MASK        0x3c6
#define PALETTE_REGISTER_RD 0x3c7
#define PALETTE_REGISTER_WR 0x3c8
#define PALETTE_DATA        0x3c9
#define SCREEN_WIDTH 320
#define SCREEN_HEIGHT 200

#define MODEY_SCREEN_WIDTH 80
#define MODEY_SCREEN_HEIGHT 200

#define MODEZ_SCREEN_WIDTH 80
#define MODEZ_SCREEN_HEIGHT 400

#define CHAR_WIDTH  8
#define CHAR_HEIGHT 8

// screen transition commands

#define SCREEN_DARKNESS  0         // fade to black
#define SCREEN_WHITENESS 1         // fade to white
#define SCREEN_WARP      2         // warp the screen image
#define SCREEN_SWIPE_X   3         // do a horizontal swipe
#define SCREEN_SWIPE_Y   4         // do a vertical swipe
#define SCREEN_DISOLVE   5         // a pixel disolve

// defines add for split screen stuff


#define VGA_INPUT_STATUS_1   0x3DA // VGA input status register number 1
                                   // D3 is vertical retrace bit
                                   // 1 - retrace in progress
                                   // 0 - no retrace

#define VGA_VRETRACE_MASK 0x08     // masks off unwanted bit of status reg

#define SPRITE_WIDTH      16       // default width of a sprite
#define SPRITE_HEIGHT     16       // default height of a sprite

#define MAX_SPRITE_FRAMES 32

// sprite states

#define SPRITE_DEAD       0
#define SPRITE_ALIVE      1
#define SPRITE_DYING      2

// M A C R O S ///////////////////////////////////////////////////////////////

#define SET_SPRITE_SIZE(w,h) {sprite_width=w; sprite_height=h;}


#define GRAPHICS_MODE13		0x13  	/* 320x200x256 */
#define TEXT_MODE			0x03  	/* 80x25 text mode */

#define COLOR_MASK			0x3C6 	// the bit mask register
#define COLOR_REGISTER_RD	0x3C7   // set read index at this I/O
#define COLOR_REGISTER_WR	0x3C8	// set write index at this I/O
#define COLOR_DATA			0x3C9	// the R/W data is here

#define ROM_CHAR_SET_SEG 	0xF000	// segment of 8x8 ROM character set
#define ROM_CHAR_SET_OFF	0xFA6E  // beginning offset of 8x8 ROM character set
#define ROM_CHAR_HEIGHT		8
#define ROM_CHAR_WIDTH		8

#define SPRITE_WIDTH      16
#define SPRITE_HEIGHT     16


#define SPRITE_DEAD       0
#define SPRITE_ALIVE      1
#define SPRITE_DYING      2

#define SEQUENCER				0x3C4	// the sequencer's index port
#define CRT_CONTROLLER			0x3D4	// the crt controller's index port
#define GFX_CONTROLLER			0x3CE	// the graphics controller's index port
#define GC_DATA_REG				0x3CF
#define ATTR_CONTROLLER_FF      0x3DA	// the attribute controller's Flip Flop
#define ATTR_CONTROLLER_DATA	0x3C0	// the attribute controller's data port

#define VGA_INPUT_STATUS_1		0x3DA   // vertical retrace flag
#define VGA_VRETRACE_MASK 		0x08    // masks off unwanted bit of status reg

#define VR_BIT                0x08
#define DISPLAY_ENABLE        0x01

// ...defines for the CRT controller registers of interest

#define CRT_OVERFLOW                    0x07    // overflow register

#define CRT_MAX_SCANLINE		0x09	// the maximum scanline register
										// used to select how many scanlines
										// per row

#define CRT_ADDR_MODE			0x14	// the address mode register
                                                // used to select byte addressing
                                                // for VGA
										// also known as the underline register

#define CRT_MODE_CONTROL		0x17    // the mode control register
		                                // used to select single byte addressing
#define  CRT_LINE_COMPARE               0x18   // the CRT controller's line
                                               // compare index
// ...defines for the GFX controller registers of interest

#define GFX_READ_PLANE_ENABLE           0x04

#define GFX_WRITE_MODE			0x05	// the memory write mode register
	                                        // used to deselect even/odd plane
	                                        // addressing

#define GFX_MISC			0x06    // the miscellaneous register
                                                // used to deselect the chaining
	                                        // of memory

#define BIT_MASK                        0x08    // the GC bitmask

// ...defines for the SEQUENCER registers of interest

#define SEQ_PLANE_ENABLE		0x02	// plane enable register, used to select
                                                // which planes are written to by a
	                                        // CPU write

#define SEQ_MEMORY_MODE			0x04    // the memory mode register
		                                // used to deselect memory chain mode
                		                // and odd/even memory addressing
#define MAP_MASK                        0x02    // the Map Mask


// M A C R O S - for use with MODE Z //////////////////////////////

#define SET_BITS(x,bits)	(x | bits) 	// used to set bits in a word
#define RESET_BITS(x,bits)  (x & ~bits) // used to reset bits in a word

// these are used to change the visual page of the VGA

#define CRT_ADDR_LOW			0x0D    // the index of the low byte of the start address
#define CRT_ADDR_HI				0x0C	// the index of the hi byte of the start address

// mode X (if other routines supported it), Y, Z page defines

#define PAGE_0					0
#define PAGE_1					1
#define PAGE_2                                  2
#define PAGE_3                                  3

// this structure holds a RGB triple in three unsigned bytes
typedef struct rgb_color_type 
        {
        unsigned char red;
        unsigned char green;
        unsigned char blue;
        } rgb_color, *rgb_color_ptr;

typedef struct rgb_palette_typ
        {
        int start_reg;   // index of the starting register that is save
        int end_reg;     // index of the ending registe that is saved

        rgb_color colors[256];   // the storage area for the palette

        } rgb_palette, *rgb_palette_ptr;

typedef struct pcx_header_typ
        {
        char manufacturer;
        char version;
        char encoding;
        char bits_per_pixel;
        int x,y;
        int width,height;
        int horz_res;
        int vert_res;
        char ega_palette[48];
        char reserved;
        char num_color_planes;
        int bytes_per_line;
        int palette_type;
        char padding[58];
        } pcx_header, *pcx_header_ptr;


typedef struct pcx_picture_typ
        {
        pcx_header header;
        rgb_color palette[256];
        char *buffer;
        } pcx_picture, *pcx_picture_ptr;


// I employed two of Andre's sprite routines... so I used his sprite
// structure - and modified it for Mode Y implementation

// I don't use all the variables... such as x_clip - although it's a good
// idea to use that, so you can save some work in the pipeline.

typedef struct sprite_typ
        {
        int x,y;            // position of sprite
        int width,height;   // dimensions of sprite in pixels

        int counter_1;      // some counters for timing and animation
        int counter_2;
        int counter_3;

        int threshold_1;    // thresholds for the counters (if needed)
        int threshold_2;
        int threshold_3;

        unsigned char  *frames[MAX_SPRITE_FRAMES]; // array of pointers to
                                                      // the images

        int curr_frame;                // current frame being displayed
        int num_frames;                // total number of frames
        int state;                     // state of sprite, alive, dead...
        unsigned char  *background; // image under the sprite

        int x_clip,y_clip;             // clipped position of sprite
        int width_clip,height_clip;    // clipped size of sprite
        int visible;                   // used by sprite engine to flag
                                       // if a sprite was invisible last
                                       // time it was drawn hence the background
                                       // need not be replaced

        } sprite, *sprite_ptr;

// this is the typedef for a bitmap

typedef struct bitmap_typ
        {
        int x,y;                    // position of bitmap
        int width,height;           // size of bitmap
        unsigned char  *buffer;  // buffer holding image

        } bitmap, *bitmap_ptr;

extern char *video_buffer;
extern char *double_buffer;
extern char *rom_char_set;

extern char *pagez_0_buffer;
extern char *pagez_1_buffer;

extern int mode_z_page;

extern int screen_width_mode;
extern int screen_height_mode;

// Function Prototypes

// Mode Z Stuff

void Fill_Screen_Y(int color);
void Fill_Screen_Z(int color);

void Write_Pixel_Z(int x, int y, int color);

void Set_Mode_Z(void);

void Set_Working_Page_Mode_Z(int page);
void Set_Visual_Page_Mode_Z(int page);

void Set_Visual_Page_Mode_Z_Half(int page);

// ROUTINES THAT ARE A LITTLE TOO SPECIFIC FOR A GENERAL PURPOSE
// LIBRARY. (HINT: USED IN THE GAMELET - "SLOW RUNNINGS"

void PCX_Copy_To_Video_Mem(pcx_picture_ptr image, unsigned char *dest);

void Copy_From_Unseen_Vram_to_Page(unsigned char *buffer1, unsigned char *buffer2, int to_row);

void Planar_Special_Fade_to(unsigned char *buffer1, unsigned char *buffer2, int to_row);

void Print_Char_Mode_Z(int xc, int yc, char c, int color, int transparent);
void Print_String_Mode_Z(int x, int y, int color, char *string, int transparent);

void Line_H_Mode_Z(int x1, int x2, int y, int color);
void Line_V_Mode_Z(int y1, int y2, int x, int color);

void Set_Palette_Register(int index, rgb_color_ptr color);

void Get_Palette_Register(int index, rgb_color_ptr color);

void PCX_Init(pcx_picture_ptr image, int width, int height);

void PCX_Load(char *filename, pcx_picture_ptr image, int width, int height, int enable_palette);

void PCX_Delete(pcx_picture_ptr image);

void PCX_Get_Sprite(pcx_picture_ptr image,
                     sprite_ptr sprite,
                     int sprite_frame,
                     int cell_x, int cell_y);

void Sprite_Init(sprite_ptr sprite,int x,int y, int width, int height,
                                   int c1,int c2,int c3,
                                   int t1,int t2,int t3);

void Sprite_Delete(sprite_ptr sprite);

void Set_Split_Screen(int line_number);

// routines basically hijacked from the book: PC GAME PROGRAMMING EXPLORER
// (Coriolis Group Books) - by Dave Roberts

void Wait_Vertical_Retrace_End(void);
void Wait_Display_Mode(void);
void Wait_Retrace_Mode(void);
int Vertical_Retrace_Occuring(void);

