// poly3z.c   Special thanks to Dhonn Lushine (dhonn@usa.net)
//                for allowing me to use his TimerDJ.h module
//
// Author: James McCue
// Author e-mail: aloiterer@juno.com
// Author homepage:
//      http://www.fortunecity.com/skyscraper/gigaherts/179/index.html
//            (jeez! - that's a long URL...)
//
//    This program and it's related resources:
//           poly3z.c            TimerDJ.h
//           magic11z.c          magic11z.h
//           modez.c             modez.h
//           input.c             input.h
//           texturez.c          texturez.h
//        ...along with any included .plgs
//
//    - are without warrantee of any kind. The author assumes no liability
//   for damage that may arise via the use, or abuse, of this code...
//
//    This program, and it's resources rely so heavily on the work of
// Andre' Lamothe in his Black Art of 3D Programming (waite group press),
// that I recommend anyone looking to REALLY understand this stuff get his
// book IMMEDIATELY!
//
// This program simply loads a modified type of .plg, and allows the user
// to rotate it around three (3) distinct axis... X, Y, and Z - and translate
// (move) the object along the z-axis
//
// This time around... the code is better... though my new rotation code
// could be much more efficient!
//
// Many of the routines used here are from chapters 11-15 of Andre's book!
//
// compile this with the line:
//   gcc -s -fomit-frame-pointer -funroll-loops -O2 poly3z.c magic11z.c
//       modez.c input.c texturez.c -o 3dtut3z.exe
//
// ... or just dump the sources into a project under RHIDE and choose:
// MAKE
//
// NOTE: It is advisable to have .plgs setup for some form of TMapping if
//   you're going to switch to Texture mapped modelling modes at some time
//   during the running of the demo...
//
// I N C L U D E S ///////////////////////////////////////////////////////////

#include <sys/nearptr.h>
#include <go32.h>
#include <dos.h>         
#include <stdlib.h>
#include <stdio.h>
#include <pc.h>
#include <dpmi.h>
#include <time.h>
#include "modez.h"
#include "input.h"
#include "magic11z.h"
#include "texturez.h"

// include Dhonn Lushine's Timer function (dhonn@usa.net)
//                                http://members.aol.com/dhonn

#include "timerdj.h"

// D E F I N E S /////////////////////////////////////////////////////////////

#define rep_movsl(src, dest, numwords) \
__asm__ __volatile__ ( \
  "cld\n\t" \
  "rep\n\t" \
  "movsl" \
  : : "S" (src), "D" (dest), "c" (numwords) \
  : "%ecx", "%esi", "%edi" )

// MACRO TO SET BITS (32-bits at a time), used in filling the double
// buffer using LONG WORDS ; ]

#define rep_stosl(value, dest, numwords) \
__asm__ __volatile__ ( \
  "cld\n\t" \
  "rep\n\t" \
  "stosl" \
  : : "a" (value), "D" (dest), "c" (numwords) \
  : "%ecx", "%edi" )

#define NUM_STARS       35

#define NUM_OBJECTS     2

#define EVADING         5
#define CHASING         7

// G L O B A L S  ////////////////////////////////////////////////////////////

object test_object[NUM_OBJECTS];   // the test object


pcx_picture image_pcx;           // general pcx image
pcx_picture background_pcx;

clock_t clock(void);
clock_t start_time, end_time, frame_time;

int x_pos[NUM_STARS];
int y_pos[NUM_STARS];
int star_color[NUM_STARS];

int star_plane[NUM_STARS];

int ang_x=1, ang_y=0, ang_z=0;
int rot_change = 0;
    

int direction_x=0;
int direction_y=0;
int direction_z=0;

int movementspeed = 0; // this is for use with Dhonn's timing paradigm

// FUNCTIONS /////////////


// initialize the "star" positions, color, and plane (used in
// creating the illusion of a 3d starfield)

void Init_Stars(void)
{
     int index;

     for (index = 0; index < NUM_STARS; index++)
     {
         x_pos[index] = rand()%320;
         y_pos[index] = rand()%400;

         star_plane[index] = rand()%3+1;

         star_color[index] = 16+(star_plane[index]*16);

     } // end for index
} // end Init_Stars

// translate the stars based on some aspect of the direction the ship
// is heading in...

void Translate_Stars(int forward)
{
     int index;
     float multiplier;

     for (index = 0; index < NUM_STARS; index++)
     {
         multiplier = 0.05*(((float)star_plane[index]*movementspeed)+1);

         if (forward)
         {
             x_pos[index]-=(int)(direction_x*multiplier);
             y_pos[index]+=(int)(direction_y*multiplier);
         }
         else
         {
             x_pos[index]+=(int)(direction_x*multiplier);
             y_pos[index]-=(int)(direction_y*multiplier);
         }

         // just wrap around stars

         if (x_pos[index]<0) x_pos[index]+=320;
         if (x_pos[index]>319) x_pos[index]-=320;
         if (y_pos[index]<0) y_pos[index]+=400;
         if (y_pos[index]>399) y_pos[index]-=400;
    } // end for index

} // end Translate_Stars

// draw the starfield

void Draw_Stars(void)
{
    int which_plane, index;

    for (index = 0; index < NUM_STARS; index++)
    {
         which_plane = x_pos[index];
    
         which_plane = which_plane & 0x03;

         outportb(SEQUENCER,SEQ_PLANE_ENABLE);
         outportb(SEQUENCER+1,0x01<<which_plane);

         video_buffer[ (y_pos[index]<<6) + (y_pos[index]<<4) + (x_pos[index]>>2) ] = star_color[index];
    } // end for index

} // end Draw_Stars

void Determine_Direction(void)
{
     direction_x = test_object[0].vertices_local[12].x;
     direction_y = test_object[0].vertices_local[12].y;
     direction_z = test_object[0].vertices_local[12].z;

} // end Determine_Direction

void Translate_Ship(int forward)
{
     // movementspeed is a global, which is continuously updated by
     // Dhonn's timer function

     float movementfactor;
     float change_in_x, change_in_y, change_in_z;

     // use movementspeed to calculate a movement factor

     movementfactor = 0.3*(float)movementspeed;

     // why do I check this?

     if (movementfactor == 0)
         return;

     change_in_x = (float)direction_x*movementfactor;
     change_in_y = (float)direction_y*movementfactor;
     change_in_z = (float)direction_z*movementfactor;

     // use components of the objects origin - nosetip vector
     // to calculate the new x,y,z coordinates for the object, and
     // it's engine exhaust ; )

     if (forward)
     {
         test_object[0].world_pos.x += (int)(change_in_x);
         test_object[0].world_pos.y += (int)(change_in_y);
         test_object[0].world_pos.z += (int)(change_in_z);
    
         test_object[1].world_pos.x += (int)(change_in_x);
         test_object[1].world_pos.y += (int)(change_in_y);
         test_object[1].world_pos.z += (int)(change_in_z);
    
         view_point.x = test_object[0].world_pos.x;
         view_point.y = test_object[0].world_pos.y;
     }
     else
     {
         test_object[0].world_pos.x -= (int)(change_in_x);
         test_object[0].world_pos.y -= (int)(change_in_y);
         test_object[0].world_pos.z -= (int)(change_in_z);
    
         test_object[1].world_pos.x -= (int)(change_in_x);
         test_object[1].world_pos.y -= (int)(change_in_y);
         test_object[1].world_pos.z -= (int)(change_in_z);
    
         view_point.x = test_object[0].world_pos.x;
         view_point.y = test_object[0].world_pos.y;
     }
} // end Translate_Ship

void Store_Original_Vert_Positions(void)
{
    int index;

    for (index = 0; index < test_object[0].num_vertices; index++)
    {
        test_object[0].vertices_local_unrotated[index].x =test_object[0].vertices_local[index].x;
        test_object[0].vertices_local_unrotated[index].y =test_object[0].vertices_local[index].y;
        test_object[0].vertices_local_unrotated[index].z =test_object[0].vertices_local[index].z;
    } // end for index

    for (index = 0; index < test_object[1].num_vertices; index++)
    {
        test_object[1].vertices_local_unrotated[index].x =test_object[1].vertices_local[index].x;
        test_object[1].vertices_local_unrotated[index].y =test_object[1].vertices_local[index].y;
        test_object[1].vertices_local_unrotated[index].z =test_object[1].vertices_local[index].z;
    } // end for index

} // end Store_Original_Vert_Positions

void Restore_Original_Vert_Positions(void)
{
    int index;

    for (index = 0; index < test_object[0].num_vertices; index++)
    {
        test_object[0].vertices_local[index].x =test_object[0].vertices_local_unrotated[index].x;
        test_object[0].vertices_local[index].y =test_object[0].vertices_local_unrotated[index].y;
        test_object[0].vertices_local[index].z =test_object[0].vertices_local_unrotated[index].z;
    } // end for index

    for (index = 0; index < test_object[1].num_vertices; index++)
    {
        test_object[1].vertices_local[index].x =test_object[1].vertices_local_unrotated[index].x;
        test_object[1].vertices_local[index].y =test_object[1].vertices_local_unrotated[index].y;
        test_object[1].vertices_local[index].z =test_object[1].vertices_local_unrotated[index].z;
    } // end for index

} // end Restore_Original_Vert_Positions

void Printamess(void)
{
    clrscr();

    printf("Welcome to the Z Mode suppliment of my\n");
    printf("DJGPP 3D polygon graphics\n");
    printf("tutorial's demo program!\n\n");
    printf("keys:\n\n");
    printf("w,e,r,s,d,f:  Object rotation keys\n\n");
    printf("1,2,3,4,5,6:  Change modelling mode\n");
    printf("        (just verts, wireframe, constant shaded,\n");
    printf("        (flat shaded, textured, textured&shaded)\n\n");
    printf("u,i: change textures in textured modes...\n\n");
    printf("c: toggle rotation mode\n");
    printf("ESC        :  Exits\n\n");
    printf("Press any key to continue\n");
} // end Printamess

// MAIN MODULE //////////////////////

void main(int argc, char **argv)
{
    int index;
    
    __dpmi_regs regs;

    int done=0, // exit flag
        num_frames=0;

    int success;
    float change_ang;

    /* disable all memory protection */
    if(success=__djgpp_nearptr_enable())
    {
        video_buffer += __djgpp_conventional_base;
        pagez_0_buffer += __djgpp_conventional_base;
        pagez_1_buffer += __djgpp_conventional_base;

        rom_char_set += __djgpp_conventional_base;
    } // if able to disable memory protection
    else
    {
        printf("For some reason, this program has been unable to disable\n");
        printf("memory protection... and as this has not been tested without\n");
        printf("disabling it, let's play it save - Exiting program\n");
        printf("\n...but still - DO look at the source, there should be something\n");
        printf("to learn by in it. Oh - and e-mail me if you get this message PLEASE!\n");

        return(0);
    }

////////////////////////////////////////////////////////

    Printamess();

    getch();

    // create the sin/cos lookup tables used for the rotation function
    
    Build_Look_Up_Tables();

    // load a plg file...

    if(!PLG_Load_Object(&test_object[0],"cube.plg",1))
    {
           printf("Error in cube.plg - press any key\n");
           getch();
           exit(0);
    }
    if(!PLG_Load_Object(&test_object[1],"flame.plg",1))
    {
           printf("Error in flame.plg - press any key\n");
           getch();
           exit(0);
    }

    Init_Stars();

    clip_near_z = 50;

    Store_Original_Vert_Positions();

    Position_Object((object_ptr)&test_object[0],0,0,300);
    Position_Object((object_ptr)&test_object[1],0,0,300);
    
    // set the viewpoint
    
    view_point.x = 0;
    view_point.y = 0;
    view_point.z = 0;
    
    // playing with the following assignment will alter the FOV (field
    // of view)

    viewing_distance = 250;

    done = 0;

    num_frames = 0;

//    if (!(double_buffer=(char *)malloc(320 * 400 + 1)))
//    {
//        printf("\nSorry... I couldn't allocate the double_buffer...\n");
//        printf("\nPress any key to exit\n");
//        getch();
//    } // end if couldn't allocate the db

    Set_Mode_Z();

    // load in the textures
    
    PCX_Load("shuttile.pcx", (pcx_picture_ptr)&image_pcx,320,200,1);

    // what I'm doing here might be a little confusing...
    // y_text_shift is a global variable used in the texture.c module...

    // I wanted you to be able to use different sized textures...
    // y_texture_shift basically allows you to use different
    // widths/heights for the sprites used in texturing...

    // It's set to use powers of two... (i.e. 64x64, 128x128, 256x256...
    // Sprite_Init hasn't changed... (1<<y_texture_shift) means
    // 128x128... [just be sure to change the uxv coords in the plg, or
    // you won't see any change...

    y_texture_shift = 7;

    Sprite_Init((sprite_ptr)&textures,0,0,(1<<y_texture_shift),
                     64,0,0,0,0,0,0);

    PCX_Get_Sprite((pcx_picture_ptr)&image_pcx, (sprite_ptr)&textures,0,0,0);
    PCX_Get_Sprite((pcx_picture_ptr)&image_pcx, (sprite_ptr)&textures,1,1,0);

    PCX_Delete((pcx_picture_ptr)&image_pcx);

    if (!Load_Bigger_Palette())
        Make_Bigger_Palette();

    Show_Bigger_Palette();

//    Set_Split_Screen(150);

    getch();

    // Big D's timer function

////////////Dhonn's timer installation code ///////////////
// lock the code and data, to prevent crashing when memory might start
// "paging" (i.e.: when virtual memory is being used!)

//    _go32_dpmi_lock_data(&timermyfunction,sizeof(timermyfunction)); /* need to lock memory because hardware interrupts do allow paging */
//    _go32_dpmi_lock_code(myfunction,(long)myfunction_end-(long)myfunction); /* need to lock code also */
    
// LOCKING is a good idea, because if we get into a situation where
// memory starts "paging" (i.e.: virtual memory is being used by swapping
// pages of memory from mem to disk & vice versa) ... timer & keyboard
// ISRs that aren't LOCKED can crash!

    _go32_dpmi_lock_data(&keys_active,sizeof(keys_active));
    _go32_dpmi_lock_code(key_handler,(long)key_handler_end-(long)key_handler);

    setbuf(stdout,NULL);		/* fix djgpp's buffered output */

    key_init();

    install_timer(NULL,TIMER_(60));		/* 60hz timer reprogram */

    which_euler_rotate_mode = 1;

    start_time = clock();  // the first call to clock always returns 0...
    //startime=floatsystime();

    test_object[0].state = EVADING;

    while(!done)
    {
        if (test_object[0].world_pos.z > 600)
            test_object[0].state = CHASING;

        if (test_object[0].world_pos.z < 150)
            test_object[0].state = EVADING;

        Wait_Retrace_Mode();

        Set_Working_Page_Mode_Z(mode_z_page);

        outportb(SEQUENCER, SEQ_PLANE_ENABLE);
        outportb(SEQUENCER+1,0x0f);

        rep_stosl(0x0000,video_buffer,8000);


        // These next 6 will dynamically change the shading type
        // for the object at runtime

        if (key_table[MAKE_1])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].shading = JUST_VERTICES;

        if (key_table[MAKE_2])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].shading = WIRE_FRAME;

        if (key_table[MAKE_3])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].shading = CONSTANT_SHADING;

        if (key_table[MAKE_4])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].shading = FLAT_SHADING;

        if (key_table[MAKE_5])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].shading = TEXTURED_POLYS;

        if (key_table[MAKE_6])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].shading = TEXTURED_SHADED;

/// change texture /////

        if (key_table[MAKE_I])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].texture_number=1;

        if (key_table[MAKE_U])
            for (index = 0; index<=(test_object[0].num_polys)-5; index++)
                test_object[0].polys[index].texture_number=0;

//////////////////////////
        if (key_table[MAKE_C])
        {
            while(key_table[MAKE_C]){};

            which_euler_rotate_mode = !which_euler_rotate_mode;

            Restore_Original_Vert_Positions();

            test_object[0].world_pos.x = 0;
            test_object[0].world_pos.y = 0;
            test_object[0].world_pos.z = 300;

            test_object[1].world_pos.x = 0;
            test_object[1].world_pos.y = 0;
            test_object[1].world_pos.z = 300;

            // reposition viewpoint

            view_point.x = 0;
            view_point.y = 0;
            view_point.z = 0;

            // not really necessary yet...

            view_angle.ang_x = 0;
            view_angle.ang_y = 0;
            view_angle.ang_z = 0;
        } // end change rotation mode

//////////////////////////
        // if no rotation keys are pressed, I want the object NOT to rotate,
        // so I reset these to 0 each time...

        if (!which_euler_rotate_mode)
        {
            ang_x = 0;
            ang_y = 0;
            ang_z = 0;
        }

        rot_change = 0;

        movementspeed=getmovementspeed();

//endtime=floatsystime();
//movementspeed=endtime-startime;

//xang+=movementspeed*angles_per_second;

        // test what key(s) user is pressing

// NOTE: the first six queries determine the continued rotation
//    of the "ship", or whatever you want to rotate... general numbers
//    past 180 are used for rotations in the OTHER direction - NOT negative
//    numbers!

        change_ang = movementspeed*5;

        if (key_table[MAKE_E] || key_table[MAKE_D] || key_table[MAKE_F] ||
            key_table[MAKE_S] || key_table[MAKE_W] || key_table[MAKE_R])
               rot_change = 1;

        if (key_table[MAKE_E])
            ang_x+=change_ang;
    
        if (key_table[MAKE_D])
            ang_x-=change_ang;
    
        if (key_table[MAKE_F])
            ang_z+=change_ang;
    
        if (key_table[MAKE_S])
            ang_z-=change_ang;
    
        if (key_table[MAKE_W])
            ang_y+=change_ang;
    
        if (key_table[MAKE_R])
            ang_y-=change_ang;

        if (ang_x>=360) ang_x = ang_x % 360;
        if (ang_y>=360) ang_y = ang_y % 360;
        if (ang_z>=360) ang_z = ang_z % 360;

        while(ang_x<0)
             ang_x += 360;
        while(ang_y<0)
            ang_y += 360;
        while(ang_z<0)
            ang_z += 360;

        // rotate the objects LOCAL vertices around the object's LOCAL
        // "center" point... depending on how you setup the plg - the
        // "center" doesn't have to be anywhere within the object...

        // rotate the object on all three axes
        // my handling of the new rotation stuff isn't very efficient,
        // so I do this to try to avoid some of the unnecessary rotations

        if (rot_change)
        {
            Rotate_Object((object_ptr)&test_object[0],ang_x,ang_y,ang_z);
            Rotate_Object((object_ptr)&test_object[1],ang_x,ang_y,ang_z);
        }

        // Do you want to move the object up and down the vector
        // between it's local origin, and the tip of it's nose??

        Determine_Direction();

        if (key_table[MAKE_UP])
        {
            Translate_Ship(1);
            Translate_Stars(1);
        }
        if (key_table[MAKE_DOWN])
        {
            Translate_Ship(0);
            Translate_Stars(0);
        }

        // test for exit key
    
        if (key_table[MAKE_ESC])
            done=1;

/////////////////////added to toggle the exhaust of the ship//////

        test_object[1].state = rand()%2;

        Draw_Stars();

        // Creating the world to camera transformation matrix isn't really
        // "necessary" yet... because there are no keys to move the viewpoint
        // or view angles yet...

        // basically what this next call does is this:
        //   No matter where the viewpoint is in the world, or how it is
        // rotated... everything we'll be doing is based on a world viewpoint
        // of 0,0,0 and view angles of 0,0,0... So based on where we are,
        // we figure out how we have to rotate all the objects in the world,
        // and move them - so that we have that 0,0,0  0,0,0 relationship..

        Create_World_To_Camera();

        // reset the poly list, (i.e.: clear it out, to assume we're not
        //                       going to see any polys <- hopefully we
        //                       will though!)
        Generate_Poly_List(NULL, RESET_POLY_LIST);

        for (index = 0; index < NUM_OBJECTS; index++)
        {
            // if the object is at least partially within the viewing fustrum
            // then...
            if (!Remove_Object(&test_object[index],OBJECT_CULL_XYZ_MODE))
            {
//////////////////////////////////////////////////////////////////////////////
        
                // convert the local coordinates into camera coordinates for projection
                // note the viewer is at (0,0,0) with angles 0,0,0 so the transformaton
                // is simply to add the world position to each local vertex
    
                Local_To_World_Object((object_ptr)&test_object[index]);
        
//////////////////////////////////////////////////////////////////////////////
    
                // At this stage there aren't any 2-sided polys... so if the
                // surface normal (picture a little arrow pointing directly
                // out of the face of the polygon) of the polygon is pointed
                // away from the light source - this next routine DOESN'T let
                // it be added to the poly list.
    
                // if it will show - calculate shading (if the poly is to have
                // FLAT_SHADING...
    
                Remove_Backfaces_And_Shade((object_ptr)&test_object[index]);
    
                // This is why we used Create_World_To_Camera, this next routine
                // uses the matrix created by that routine to give all points
                // RELATIVE to the viewpoint & viewangles (as though we were at
                // 0,0,0 with view angles 0,0,0...
    
                World_To_Camera_Object((object_ptr)&test_object[index]);
        
                // This routine (the way it's called here) will "kill" polys
                // that are either to far in front of us, or behind us...
    
                Clip_Object_3D((object_ptr)&test_object[index],CLIP_XYZ_MODE);
    
                // Generate_Poly_List again - this time to add a poly to the list
                // IF the poly hasn't been "killed" yet...
    
                Generate_Poly_List(&test_object[index],ADD_TO_POLY_LIST);
            }
        }
        // Does what it says (using qsort())
        Sort_Poly_List();

        // Well... this doesn't draw ANY poly list yet...

        Draw_Any_Poly_List_Planar();

        // move double buffer to unseen page

/*
        outportb(SEQUENCER, SEQ_PLANE_ENABLE);
        outportb(SEQUENCER+1, 0x01);
        rep_movsl(double_buffer, video_buffer, 8000);
        outportb(SEQUENCER, SEQ_PLANE_ENABLE);
        outportb(SEQUENCER+1, 0x02);
        rep_movsl(double_buffer+32000, video_buffer, 8000);
        outportb(SEQUENCER, SEQ_PLANE_ENABLE);
        outportb(SEQUENCER+1, 0x04);
        rep_movsl(double_buffer+64000, video_buffer, 8000);
        outportb(SEQUENCER, SEQ_PLANE_ENABLE);
        outportb(SEQUENCER+1, 0x08);
        rep_movsl(double_buffer+96000, video_buffer, 8000);
                                                    */

        // display double buffer

        Wait_Display_Mode();     // return to calling function (this one)
                                 // if the element is currently scanning
                                 // pixels... (i.e.: not in retrace)

        Set_Visual_Page_Mode_Z(mode_z_page);

        mode_z_page = !mode_z_page;
        num_frames++;
    
        // lock onto 18 frames per second max
    
    } // end while

    // uninstall big D's timer!

    uninstall_timer();	/* uninstall timer now so I can start a new one */

    end_time=clock();
    
    /* delete keyboard handler */
    key_delete();
    
    /* go back to text mode */
    regs.x.ax = 0x03; /* 0x03 is the mode number */
    __dpmi_int(0x10, &regs);

    PCX_Delete((pcx_picture_ptr)&background_pcx);

    Sprite_Delete((sprite_ptr)&textures);

    for (index = 0; index < NUM_OBJECTS; index++)
        free((object_ptr)&test_object[index]);

    /* reenable memory protection */
    __djgpp_nearptr_disable();
    
    printf("%d seconds have elapsed\n", (end_time-start_time)/CLOCKS_PER_SEC);
    printf("%f frames per second...\n", (float)num_frames/(float)((end_time-start_time)/CLOCKS_PER_SEC));
    
    getch();
    
    exit(0);

} // end main
