// texturez.c
//
// Affine texture-mapping and more..
//
//   The texture mapping uses
// an expanded color look-up table to simulate more than modez's limiting
// 256 colors.
//
//   Special thanks to Dhonn Lushine (dhonn@usa.net)
//                        http://members.aol.com/dhonn
//    - for teaching me affine TMapping.. and about Clut palettes... and
//  always arguing with me!!
//
//    Other routines in here are for creating, saving, loading, and
// displaying the Clut (expanded) palette we'll use for doing
// shaded texture-mapping. These routines could also be used with Gouraud
// shaded textures, and even phong...
//
//         - Jim the loiterer
//          aloiterer@juno.com
//          http://www.fortunecity.com/skyscraper/gigahertz/179/index.html

#include <stdio.h>
#include <sys/nearptr.h>
#include <go32.h>
#include <dos.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <pc.h>
#include <dpmi.h>
#include <search.h>				// this one is need for qsort()

#include "modez.h"
#include "input.h"
#include "magic11z.h"
#include "texturez.h"

int color_table[256][64];

int y_texture_shift = 6; // default this value to 6
                         // so we'll assume 64x64 sprites are being used

void (*rend_func_ptr) (int column, int e_y1, int e_y2, int e_u1,
                                int e_u2, int e_v1, int e_v2, int text_shade);

// I pass ALOT of parameters to these routines - I know...

void Draw_Triangle_Planar_ShadeText(int x1, int y1, int x2, int y2, int x3, int y3,
                 int t1x, int t1y, int t2x, int t2y, int t3x, int t3y, int text_shade)
{
	// this function draw a textured-triangle on the screen,
        // AND is set to use flat shading (thanks to a clut palette
        // we've set up...

        int     right_1,                // distance from top to middle of triangle on y axis
		right_2;		// distance from middle to bottom triangle on y axis


        float width;           // the of the triangle (or a subdivision of it)
              
        int column;          // looping variables

        // the following floats are used for keeping track of where the
        // edges of the triange, and the texture being used are...

        float edge_y1,edge_y2;
        float edge_u1,edge_u2,edge_v1,edge_v2;

        // these are to store the incremental changes in the edges
        // of the triangle AND the area of a given texture where
        // pixels are being sampled from...

        float dx_u1,dx_u2,dx_v1,dx_v2,dx_y1,dx_y2;

        // current UxV coordinates in a given texture

        int u1,u2,u3,v1,v2,v3;

        // used in sorting texture coordinates

        int temp_x, temp_y;

        int min_x, max_x;

        // assign the texture to a pointer

// lay out uv coordinates (step in a half step...)

    u1 = t1x+0.5;
    v1 = t1y+0.5;
    u2 = t2x+0.5;
    v2 = t2y+0.5;
    u3 = t3x+0.5;
    v3 = t3y+0.5;

// sort vertices by x values...

    if (x2<x1)	// sort p1,p2,p3 in ascending x order	if (x2<x1)
    {
        temp_x = x2;
        temp_y = y2;
        x2 	   = x1;
        y2     = y1;
        x1     = temp_x;
        y1     = temp_y;
        temp_x = u2;
        temp_y = v2;
        u2 = u1;
        v2 = v1;
        u1 = temp_x;
        v1 = temp_y;

    } // end if

    // now we know that p1 and p2 are in order

    if (x3<x1)
    {
        temp_x = x3;
        temp_y = y3;
        x3 	   = x1;
        y3     = y1;
        x1     = temp_x;
        y1     = temp_y;
        temp_x = u3;
        temp_y = v3;
        u3 = u1;
        v3 = v1;
        u1 = temp_x;
        v1 = temp_y;


    } // end if

    // finally test x3 against x2

    if (x3<x2)
    {
        temp_x = x3;
        temp_y = y3;
        x3     = x2;
        y3     = y2;
        x2     = temp_x;
        y2     = temp_y;
        temp_x = u3;
        temp_y = v3;
        u3     = u2;
        v3     = v2;
        u2     = temp_x;
        v2     = temp_y;

    } // end if


    width = x3-x1;
    if(!width) return; // this might be redundant, but we won't draw
                       // vertical lines

    // compute the change in y, u, and z for the longest side of the
    // triangle (with respect to x)

    dx_y1=(float)(y3-y1)/width;
    dx_u1=(float)(u3-u1)/width;
    dx_v1=(float)(v3-v1)/width;

    // setup starting coords (world and UV)

    edge_y1=y1;
    edge_u1=u1;
    edge_v1=v1;


    right_1 = x2;       // point where the triangle may be split horizontally
                        // to make 2 triangles...

    right_2 = x3;       // the rightmost point of the triangle


    width = x2-x1;

    if (width) // then draw left pointing triangle
    {
        dx_y2=(float)(y2-y1)/width;  // change in y
        dx_u2=(float)(u2-u1)/width;  // change in u (texture space)
        dx_v2=(float)(v2-v1)/width;  // change in v (texture space)

        edge_y2=edge_y1;  // if we're starting with the left side
        edge_u2=edge_u1;  // of the triangle, we're starting with the
        edge_v2=edge_v1;  // same point!

        // compute starting texture coordinates

        // perform x clipping

        if (x1<poly_clip_min_x)
        {
            if (x2<poly_clip_min_x) // if the second point is ALSO off
            {                       // the left hand side of the screen
                edge_y1+=dx_y1 * (float)(-x1+x2);   // this adjust for it...
                edge_y2+=dx_y2 * (float)(-x1+x2);
                edge_u1+=dx_u1 * (float)(-x1+x2);
                edge_u2+=dx_u2 * (float)(-x1+x2);
                edge_v1+=dx_v1 * (float)(-x1+x2);
                edge_v2+=dx_v2 * (float)(-x1+x2);
            }
            else
            {
                edge_y1+=dx_y1 * (float)(-x1);
                edge_y2+=dx_y2 * (float)(-x1);
                edge_u1+=dx_u1 * (float)(-x1);
                edge_u2+=dx_u2 * (float)(-x1);
                edge_v1+=dx_v1 * (float)(-x1);
                edge_v2+=dx_v2 * (float)(-x1);
            }
            min_x = 0;
        }
        else
            min_x=x1;

        if (right_1>poly_clip_max_x)
            max_x=poly_clip_max_x;
        else
            max_x=right_1;

        // test if y clipping is needed

        if (y1>=poly_clip_min_y && y1<=poly_clip_max_y &&
            y2>=poly_clip_min_y && y2<=poly_clip_max_y &&
            y3>=poly_clip_min_y && y3<=poly_clip_max_y)
        {
            rend_func_ptr = Vert_Scan_Text_Planar_NoClip;
        }
        else
            rend_func_ptr = Vert_Scan_Text_Planar_Clip;

        // draw a vertical - textured - strip

        for (temp_x=min_x; temp_x<max_x; temp_x++)
        {
            // start off working texture coordinates for current row


            column = temp_x; // leftover from planar stuff...

            (*rend_func_ptr) (column, edge_y1, edge_y2, edge_u1,
                                  edge_u2, edge_v1, edge_v2, text_shade);


            // step along 2 edges, both in screen coordinates
            // AND texture coordinates

            edge_y1+=dx_y1;
            edge_y2+=dx_y2;
            edge_u1+=dx_u1;
            edge_u2+=dx_u2;
            edge_v1+=dx_v1;
            edge_v2+=dx_v2;

        } // end for

    } // end drawing left side of triangle
    else
    {   // since we didn't start with a left pointing triangle
        edge_y2=y2;    // we initialize the second edges to
        edge_u2=u2;    // point #2
        edge_v2=v2;
    }

    width = x3-x2;

    if (width)
    {
        // get changes in y, u, and v with respect to x

        dx_y2=(float)(y3-y2)/width;
        dx_u2=(float)(u3-u2)/width;
        dx_v2=(float)(v3-v2)/width;

        // perform x clipping

        if (x2<poly_clip_min_x)
        {
            if (x3<poly_clip_min_x) return;

            edge_y1+=dx_y1 * (float)(-x2);
            edge_y2+=dx_y2 * (float)(-x2);
            edge_u1+=dx_u1 * (float)(-x2);
            edge_u2+=dx_u2 * (float)(-x2);
            edge_v1+=dx_v1 * (float)(-x2);
            edge_v2+=dx_v2 * (float)(-x2);

            min_x = 0;
        }
        else
            min_x=x2;

        if (right_2>poly_clip_max_x)
            max_x=poly_clip_max_x;
        else
            max_x=right_2;


        // test if y clipping is needed

        if (y1>=poly_clip_min_y && y1<=poly_clip_max_y &&
            y2>=poly_clip_min_y && y2<=poly_clip_max_y &&
            y3>=poly_clip_min_y && y3<=poly_clip_max_y)
        {
            rend_func_ptr = Vert_Scan_Text_Planar_NoClip;
        }
        else
            rend_func_ptr = Vert_Scan_Text_Planar_Clip;

        // draw the triangle

        for (temp_x=min_x; temp_x<=max_x; temp_x++)
        {
            column = temp_x;   // leftover from planar stuff

            // start off working texture coordinates for current column

            (*rend_func_ptr) (column, edge_y1, edge_y2, edge_u1,
                                  edge_u2, edge_v1, edge_v2, text_shade);

            // move to the next steps in the top and bottom edges
            // of the triangle, and the u,v coords.. with respect to X

            edge_y1+=dx_y1;
            edge_y2+=dx_y2;
            edge_u1+=dx_u1;
            edge_u2+=dx_u2;
            edge_v1+=dx_v1;
            edge_v2+=dx_v2;
        } // end for

    } // end if draw right side of triangle
} // end Draw_Triangle_Planar_ShadeText

// this routine draws our textured V lines

void Vert_Scan_Text_Planar_Clip(int column, int e_y1, int e_y2, int e_u1,
                                int e_u2, int e_v1, int e_v2, int text_shade)
{
    int pu1,pu2,pv1,pv2,py1,py2;

    int i;    // looping variable
    float linewidth; // used in the final drawing of a vertical strip

    long  uwidth,
          vwidth;

    long du,dv,u_curr,v_curr;	// the current u,v texture coordinates
    
    unsigned char color_me;
    
    int     x_index,   // looping variables
    	    y_index;
    int vert_strip;          // looping variables

    unsigned char *dest_addr;

    unsigned char *text; // texture pointer

    int planar_offset;

    text = textures.frames[textures.curr_frame];

    // start off working texture coordinates for current row
    
    py1 = e_y1;
    py2 = e_y2;
    pu1 = e_u1;
    pu2 = e_u2;
    pv1 = e_v1;
    pv2 = e_v2;
    
    // sort edges...
    if(py2<py1)
    {
        i=py1;
        py1=py2;
        py2=i;
    
        i=pu1;
        pu1=pu2;
        pu2=i;
    
        i=pv1;
        pv1=pv2;
        pv2=i;
    }
    
    // assign starting UxV coordinates
    
    u_curr = pu1<<16;
    v_curr = pv1<<16;
    
    linewidth=py2-py1;
    uwidth=pu2-pu1;
    vwidth=pv2-pv1;
    
    // calculate du,dv with respect to y
    
    du=(uwidth<<16)/(linewidth+1);
    dv=(vwidth<<16)/(linewidth+1);
    
    // clip texture as needed
    
    if (py1<0)
    {
        u_curr += (du * (float)(-py1));
        v_curr += (dv * (float)(-py1));
    
        py1 = 0;
    }
    
    if (py2 >=screen_height) py2 = (screen_height-1);
    
    // compute offset in the double buffer

//    planar_offset = (column & 0x03) * 16000;
    planar_offset = (column & 0x03);

    outportb(SEQUENCER,SEQ_PLANE_ENABLE);
    outportb(SEQUENCER+1,(0x01<<planar_offset));

    dest_addr = video_buffer+((int)py1<<6)+((int)py1<<4)+(column>>2);
    
    // render a verticle strip already!
    for (vert_strip=(int)py1; vert_strip <= (int)py2; vert_strip++)
    {
        // extract coordinates where a color will be copied
        // from in the texture...
    
        x_index = (int)(u_curr>>16);
        y_index = (int)(v_curr>>16);

       // extract pixel

        color_me = text[(y_index<<y_texture_shift) + x_index];
    
    // but this time, use it as part of an index into our
    // new color_table
    
        color_me = color_table[(int)color_me][text_shade];
    
        // plot the point
    
        *dest_addr = color_me;
    
        // move one line down in the video_buffer
    
        dest_addr += 80;
    
        // move to the next next (u,v) coordinate for the
        // texture to be used in this strip
    
        u_curr+=du;
        v_curr+=dv;
    }
} // end Vert_Scan_Text_Planar_Clip

void Vert_Scan_Text_Planar_NoClip(int column, int e_y1, int e_y2, int e_u1,
                                int e_u2, int e_v1, int e_v2, int text_shade)
{
    int pu1,pu2,pv1,pv2,py1,py2;

    int i;    // looping variable
    float linewidth; // used in the final drawing of a vertical strip

    long  uwidth,
          vwidth;

    long du,dv,u_curr,v_curr;	// the current u,v texture coordinates
    
    unsigned char color_me;

    int     x_index,   // looping variables
    	    y_index;
    int vert_strip;          // looping variables
    
    unsigned char *dest_addr;

    unsigned char *text; // texture pointer

    int planar_offset;

    text = textures.frames[textures.curr_frame];

    // start off working texture coordinates for current row
    
    py1 = e_y1;
    py2 = e_y2;
    pu1 = e_u1;
    pu2 = e_u2;
    pv1 = e_v1;
    pv2 = e_v2;
    
    // sort edges...
    if(py2<py1)
    {
        i=py1;
        py1=py2;
        py2=i;
    
        i=pu1;
        pu1=pu2;
        pu2=i;
    
        i=pv1;
        pv1=pv2;
        pv2=i;
    }
    
    // assign starting UxV coordinates
    
    u_curr = pu1<<16;
    v_curr = pv1<<16;
    
    linewidth=py2-py1;
    uwidth=pu2-pu1;
    vwidth=pv2-pv1;
    
    // calculate du,dv with respect to y
    
    du=(uwidth<<16)/(linewidth+1);
    dv=(vwidth<<16)/(linewidth+1);
    
    // compute offset in the double buffer
    
    planar_offset = (column & 0x03);

    outportb(SEQUENCER,SEQ_PLANE_ENABLE);
    outportb(SEQUENCER+1,(0x01<<planar_offset));

    dest_addr = video_buffer+((int)py1<<6)+((int)py1<<4)+(column>>2);

    // render a verticle strip already!
    for (vert_strip=(int)py1; vert_strip <= (int)py2; vert_strip++)
    {
        // extract coordinates where a color will be copied
        // from in the texture...
    
        x_index = (int)(u_curr>>16);
        y_index = (int)(v_curr>>16);
    
        // extract pixel

        color_me = text[(y_index<<y_texture_shift) + x_index];

    // but this time, use it as part of an index into our
    // new color_table
    
        color_me = color_table[(int)color_me][text_shade];
    
        // plot the point
    
        *dest_addr = color_me;
    
        // move one line down in the video_buffer
    
        dest_addr += 80;
    
        // move to the next next (u,v) coordinate for the
        // texture to be used in this strip
    
        u_curr+=du;
        v_curr+=dv;
    }
} // end Vert_Scan_Text_Planar_NoClip

// Here is where we make our enhanced palette...

// The limitation in mode13h is having 256 simultaneous colors, and what
// making a "clut" palette (which is what the following routine does), is
// to try to simulate more colors

void Make_Bigger_Palette(void)
{
     rgb_palette a_palette;

     int index1, index2;
     float red, green, blue;
     int send_red, send_green, send_blue;
     char ch_buffer[80];
     float ri, gi, bi;         // increments for colors

     // first, since I don't have a full palette struct floating around
     // in memory, I'll load this one with the current palette

     for (index1=0; index1<256; index1++)
         Get_Palette_Register(index1, (rgb_color_ptr)&a_palette.colors[index1]);

     Print_String_Mode_Z(10,10,170,"examining current palette",0);

     // loop through all colors

     for (index1=0; index1<256; index1++)
     {
         // extract red, green, and blue factors...

         red=(float)a_palette.colors[index1].red;
         green=(float)a_palette.colors[index1].green;
         blue=(float)a_palette.colors[index1].blue;

         // double the intensities

         red*=2;
         green*=2;
         blue*=2;

         // get the change in red, green, and blue

         ri=red/64.0;
         gi=green/64.0;
         bi=blue/64.0;

         // loop through shades we'd like to have
         // we'll start with the 16th position of the table being
         // the highest intensity

         for(index2=63;index2>=0;index2--)
         {
             send_red = red;
             send_green = green;
             send_blue = blue;

             // since 63 would be the maximum if we were
             // talking the actual color palette... we're just testing
             // boundaries here

             if (send_red>63) send_red = 63;
             if (send_green>63) send_green = 63;
             if (send_blue>63) send_blue = 63;

             // call another function to get the color that best fits
             // a shade we would like

             color_table[index1][index2]=Get_Best_Fit_Color(send_red,send_green,send_blue,(rgb_palette_ptr)&a_palette);

             // decrement red, green, and blue factors

             red-=ri;
             green-=gi;
             blue-=bi;
         }
    }

    // Save the Bigger Palette to the disk so we don't have to wait
    // for the table to be created next time

    Save_Bigger_Palette();
} // end Make_Bigger_Palette

int Get_Best_Fit_Color(int redy, int greeny, int bluey, rgb_palette_ptr pal)
{
    int i;
    int red_dist, green_dist, blue_dist;
    long dist_new, dist_best,color_best;
    int r, g, b;
    char ch_buffer[80];

    dist_best = 100000000;

    // loop through all colors... ::sigh:: again...

    for(i=0;i<256;i++)
    {
        // get the current rgbs for a color

        r = pal->colors[i].red;
        g = pal->colors[i].green;
        b = pal->colors[i].blue;

        // these "distances" will be used to determine what color
        // would be the best match, in lieu of the color we'd really
        // like to have

        red_dist=redy-r;
        green_dist=greeny-g;
        blue_dist=bluey-b;

        // squaring and adding it all together countermands any
        // negative signs we might have run into... hopefully
        // the current palette contains a color that's pretty close...

        dist_new=red_dist*red_dist+green_dist*green_dist+blue_dist*blue_dist;

        // is it close?

        if(dist_new<=dist_best)
        {
            sprintf(ch_buffer,"making progress...");
            Print_String_Mode_Z(10,30,rand()%255,ch_buffer,0);

            color_best=i;

            dist_best=dist_new;

        }
    }

    // send the closest color back to the calling procedure

    return(color_best);

} // end Get_Best_Fit_Color

void Save_Bigger_Palette(void)
{
     // saves our clut palette to a disk

     FILE *fp, *fopen();
     int index,shade,to_be_written;

     // open a file for writing (whether it exists or not)

     fp = fopen("256x64.bal","wb");

     // loop through all "shades" of all colors, and write them to disk!

     for (index=0; index<256; index++)
     {
         for (shade=0; shade<64; shade++)
         {
             to_be_written = color_table[index][shade];

             fprintf(fp,"\n%d", to_be_written);
         }
     }

     // close the file...

     fclose(fp);
} // end Save_Bigger_Palette

int Load_Bigger_Palette(void)
{
     // Loads our clut palette, if it exists

     FILE *fp, *fopen();
     int index,shade,to_be_read;

     // open a file for reading

     if(!(fp = fopen("256x64.bal","rb")) )
          return(0);

     // read the colors of the disk, and put them into the color table

     for (index=0; index<256; index++)
     {
         for (shade=0; shade<64; shade++)
         {
             fscanf(fp, "%d", &to_be_read);
             color_table[index][shade] = to_be_read;
         }
     }

     // close the file...

     fclose(fp);

     return(1);
} // end Load_Bigger_Palette

void Show_Bigger_Palette(void)
{
     // this routine shows the real palette on a single line,
     // and then shows our enhanced palette

     int index1,index2,offset;
     int planar_offset;

     Print_String_Mode_Z(10,40,170,"I'll need to start using Dhonn's",0);
     Print_String_Mode_Z(10,50,170,"super hi-rez timer somewhere",0);
     Print_String_Mode_Z(10,60,170,"down the line...(it's not here)",0);

     Print_String_Mode_Z(10,80,170,"original palette",0);

     for (index1=0; index1<256; index1++)
     {
         planar_offset = (index1 & 0x03);

         outportb(SEQUENCER,SEQ_PLANE_ENABLE);
         outportb(SEQUENCER+1,(0x01<<planar_offset));

         video_buffer[(100<<6) + (100<<4) + (index1>>2)] = (unsigned char)index1;
     }
     Print_String_Mode_Z(10,110,170,"bigger palette",0);

     for (index1=0; index1<256; index1++)
     {
         planar_offset = (index1 & 0x03);

         outportb(SEQUENCER,SEQ_PLANE_ENABLE);
         outportb(SEQUENCER+1,(0x01<<planar_offset));

         video_buffer[(130<<6) + (130<<4) + (index1>>2)] = (unsigned char)index1;

         for (index2=0; index2<64; index2++)
         {
              offset = index2 + 130;

              video_buffer[(offset<<6) + (offset<<4) + (index1>>2)] =
                                 (unsigned char)color_table[index1][index2];
         }
     }
} // end Show_Bigger_Palette
